%global upstreamname composable_kernel
%global rocm_release 6.3
%global rocm_patch 1
%global rocm_version %{rocm_release}.%{rocm_patch}

%global toolchain rocm
# hipcc does not support some clang flags
# build_cxxflags does not honor CMAKE_BUILD_TYPE, strip out -g
#
# Problem with -Wp,-D_GLIBCXX_ASSERTIONS, strip it out
# https://github.com/ROCm/composable_kernel/issues/1405
%global build_cxxflags %(echo %{optflags} | sed -e 's/-fstack-protector-strong/-Xarch_host -fstack-protector-strong/' -e 's/-fcf-protection/-Xarch_host -fcf-protection/' -e 's/-g / /' -e 's/-Wp,-D_GLIBCXX_ASSERTIONS//' )

%bcond_with debug
%if %{with debug}
%global build_type DEBUG
%else
%global build_type RELEASE
%global debug_package %{nil}
%endif

%bcond_with export
%if %{with export}
%global build_compile_db ON
%else
%global build_compile_db OFF
%endif

# Reduce link pressure
%global _rocm_lto_cflags %{nil}

# Compression type and level for source/binary package payloads.
#  "w7T0.xzdio"         xz level 7 using %%{getncpus} threads
%define _source_payload         w7T0.xzdio
%define _binary_payload         w7T0.xzdio

Name:           composable_kernel
Version:        %{rocm_version}
Release:        1%{?dist}
Summary:        Performance Portable Programming Model for Machine Learning Tensor Operators
Url:            https://github.com/ROCm/composable_kernel
License:        MIT

Source0:        %{url}/archive/rocm-%{version}.tar.gz#/%{upstreamname}-%{version}.tar.gz

BuildRequires:  cmake
BuildRequires:  fdupes
BuildRequires:  git
BuildRequires:  ninja-build
BuildRequires:  rocm-cmake
BuildRequires:  rocm-comgr-devel
BuildRequires:  rocm-hip-devel
BuildRequires:  rocm-rpm-macros
BuildRequires:  rocm-runtime-devel

# Only x86_64 works right now:
ExclusiveArch:  x86_64

%description
Composable Kernel (CK) library aims to provide a programming
model for writing performance critical kernels for machine
learning workloads across multiple architectures including
GPUs, CPUs, etc, through general purpose kernel languages,
like HIP C++.

CK utilizes two concepts to achieve performance portability
and code maintainability:

- A tile-based programming model
- Algorithm complexity reduction for complex ML operators,
  using innovative technique we call "Tensor Coordinate
  Transformation".

%package devel
Summary: Libraries and headers for %{name}
Requires:       %{name}%{?_isa} = %{version}-%{release}

%description devel
%{summary}

%prep
%autosetup -p1 -n %{upstreamname}-rocm-%{version}

# do not error on warnings
sed -i -e 's@add_compile_options(-Werror)@#add_compile_options(-Werror)@' CMakeLists.txt
sed -i -e '/-Werror/d' cmake/EnableCompilerWarnings.cmake
# No extra warnings
sed -i -e 's@add_compile_options(-Weverything)@#add_compile_options(-Weverything)@' CMakeLists.txt
sed -i -e '/-Wextra/d' cmake/EnableCompilerWarnings.cmake
sed -i -e '/-Wunused/d' cmake/EnableCompilerWarnings.cmake
sed -i -e '/-Weverything/d' cmake/EnableCompilerWarnings.cmake
sed -i -e 's@-Wno-unknown-warning-option@-Wno-unknown-warning-option -Wno-unused-parameter@' cmake/EnableCompilerWarnings.cmake

# Do not time the kernel loading
# https://github.com/ROCm/composable_kernel/issues/1780
sed -i -e 's@CK_TIME_KERNEL 1@CK_TIME_KERNEL 0@' include/ck/ck.hpp

# No way easy way to turn off examples building
# https://github.com/ROCm/composable_kernel/issues/1781
sed -i -e 's@add_subdirectory(example)@#add_subdirectory(example)@' CMakeLists.txt
# No profiler
sed -i -e 's@add_subdirectory(profiler)@#add_subdirectory(profiler)@' CMakeLists.txt

# Does not honor BUILD_SHARED
# https://github.com/ROCm/composable_kernel/issues/1782
sed -i -e 's@STATIC@SHARED@' library/src/utility/CMakeLists.txt library/src/tensor_operation_instance/gpu/CMakeLists.txt
sed -i -e 's@POSITION_INDEPENDENT_CODE ON@POSITION_INDEPENDENT_CODE ON SOVERSION \"%{version}\"@' library/src/utility/CMakeLists.txt library/src/tensor_operation_instance/gpu/CMakeLists.txt

%build

# Real cores, No hyperthreading
COMPILE_JOBS=`cat /proc/cpuinfo | grep -m 1 'cpu cores' | awk '{ print $4 }'`
if [ ${COMPILE_JOBS}x = x ]; then
    COMPILE_JOBS=1
fi
# Try again..
if [ ${COMPILE_JOBS} = 1 ]; then
    COMPILE_JOBS=`lscpu | grep '^CPU(s)' | awk '{ print $2 }'`
    if [ ${COMPILE_JOBS}x = x ]; then
        COMPILE_JOBS=4
    fi
fi

# Take into account memmory usage per core, do not thrash real memory
BUILD_MEM=6
MEM_KB=0
MEM_KB=`cat /proc/meminfo | grep MemTotal | awk '{ print $2 }'`
MEM_MB=`eval "expr ${MEM_KB} / 1024"`
MEM_GB=`eval "expr ${MEM_MB} / 1024"`
COMPILE_JOBS_MEM=`eval "expr 1 + ${MEM_GB} / ${BUILD_MEM}"`
if [ "$COMPILE_JOBS_MEM" -lt "$COMPILE_JOBS" ]; then
    COMPILE_JOBS=$COMPILE_JOBS_MEM
fi

LINK_MEM=16
LINK_JOBS=`eval "expr 1 + ${MEM_GB} / ${LINK_MEM}"`

export HIP_PATH=`hipconfig -p`
clang_path=`hipconfig -l`/clang++

%cmake -G Ninja \
    -DBUILD_TESTING=OFF \
    -DCK_PARALLEL_COMPILE_JOBS=${COMPILE_JOBS} \
    -DCK_PARALLEL_LINK_JOBS=${LINK_JOBS} \
    -DCMAKE_BUILD_TYPE=%{build_type} \
    -DCMAKE_CXX_COMPILER=${clang_path} \
    -DCMAKE_INSTALL_LIBDIR=%{_libdir} \
    -DCMAKE_EXPORT_COMPILE_COMMANDS=%{build_compile_db} \
    -DCMAKE_HIP_COMPILER=${clang_path} \
    -DHIP_PLATFORM=amd \
    -DROCM_SYMLINK_LIBS=OFF

%cmake_build

%install
%cmake_install

# Need to manually strip
strip %{buildroot}/usr/lib64/lib*

# Clean up dupes
%fdupes %{buildroot}%{_prefix}

%files
%license LICENSE
%{_libdir}/libdevice_*.so.*
%{_libdir}/libutility.so.*
%exclude %{_docdir}/composablekernel/LICENSE

%files devel
%doc README.md
%{_includedir}/ck/
%{_includedir}/ck_tile/
%{_libdir}/cmake/%{name}/
%{_libdir}/libdevice_*.so
%{_libdir}/libutility.so

%changelog
* Thu Dec 26 2024 Tom Rix <Tom.Rix@amd.com> - 6.3.1-1
- Initial Package
